/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

%{
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include "is_types.h"
#include "is_lex.h"
#include "isam.h"

#ifndef lint
static	char	sccsid[] = "@(#)dbsyacc.y	1.8	09:27:08	3/30/92";
#endif

extern	int	yylineno;
extern	char	yytext[];
extern	char	last_name[];

#ifdef BSD
#define memcpy(a, b, c) bcopy((b), (a), (c))

yyerror (s)
char	*s;
{
	fprintf(stderr, "Line %d: %s\n", __LINE__, s);
}
#else
yyerror (s)
char	*s;
{
	fprintf (stderr, "line %d: %s\n", yylineno, s);
}
#endif /* BSD */

int	char_length;
int	field_type;
char	database_name[MAXFILENAME+1];
char	file_name[MAXFILENAME+1];
char	field_name[MAXFIELDNAME+1];
char	member_name[MAXFIELDNAME+1];
int	member_order;
int	index_size;
int	index_unique;
short	index_members[MAXSORTFIELDS];
dbm_t	*dbm;
dbm_t	*db_open(),
	*db_create();
int	db_open_file(),
	db_create_file(),
	db_create_index();
int	file_number;
int	field_number;

#ifdef	YYDEBUG
#	if	YYDEBUG == 1
	int	yydebug = 1;
#	endif
#endif

void
cleanup (status)
int	status;
{
	if (status) {

		/*
		 * Some syntax error in the current file, so remove
		 * it.
		 */

		if (file_name[0])
			(void) db_delete_file (dbm, file_name);
	}
	exit (status);
}
%}
%token L_DATABASE L_FILET L_FIELD L_NAME L_TYPE L_CHAR L_LENGTH L_NUMBER
%token L_SHORT L_INT L_LONG L_SERIAL L_DATE L_FLOAT L_DOUBLE L_MONEY L_INDEX
%token L_END L_UNIQUE L_OPAR L_CPAR L_SCREEN L_JOINING L_ASCENDING L_DESCENDING
%start schema
%%
database: L_DATABASE L_NAME
	{
		int	old_bsize;
		extern	int	is_blocksize;

		/*
		 * Create dictionaries with 512 byte blocks.  No point in
		 * wasting space for them ...
		 */

		old_bsize = is_blocksize;
		is_blocksize = 512;

		strncpy (database_name, last_name, MAXFILENAME);
		database_name[MAXFILENAME] = '\0';
		if (! (dbm = db_open (database_name, O_RDWR)) &&
		    ! (dbm = db_create (database_name))) {
			fprintf (stderr, "line %d: can't create database %s\n",
				yylineno, database_name);
			db_error (database_name);
			exit (1);
		}
		is_blocksize = old_bsize;
	}
	;

file:	L_FILET L_NAME
	{
		strncpy (file_name, last_name, MAXFILENAME);
		file_name[MAXFILENAME] = '\0';

		if ((file_number = db_open_file (dbm, file_name)) >= 0) {
			fprintf (stderr, "line %d: file %s exists\n",
				yylineno, file_name);
			exit (1);
		}
		if ((file_number = db_create_file (dbm, file_name)) < 0) {
			fprintf (stderr, "line %d: can't create file %s\n",
				yylineno, file_name);
			db_error (file_name);
			exit (1);
		}
	} |
	L_FILET error {
		fprintf (stderr, "line %d: expected file name, got `%s'\n",
			yylineno, yytext);
		exit (1);
	} |
	error {
		fprintf (stderr, "line %d: expected FILE, got `%s'\n",
			yylineno, yytext);
		exit (1);
	}
	;

member: L_NAME {
		strncpy (member_name, last_name, MAXFIELDNAME);
		member_name[MAXFIELDNAME] = '\0';
		member_order = L_ASCENDING;
	} |
	L_NAME L_ASCENDING {
		strncpy (member_name, last_name, MAXFIELDNAME);
		member_name[MAXFIELDNAME] = '\0';
		member_order = L_ASCENDING;
	} |
	L_NAME L_DESCENDING {
		strncpy (member_name, last_name, MAXFIELDNAME);
		member_name[MAXFIELDNAME] = '\0';
		member_order = L_DESCENDING;
	} |
	error {
		fprintf (stderr, "line %d: expected member\n", yylineno);
		cleanup (1);
	}

members: member 
	{
		int	i;
		schema_t *schema = dbm->d_files[file_number].f_schema;
		field_t	*fields = schema->s_fields;
		int	cnt = schema->s_field_count;

		index_size = 0;
		for (i = 1;i < cnt;i++) {
			if (strncmp (member_name, fields[i].f_name,
					MAXFIELDNAME) == 0) {
				if (member_order == L_DESCENDING)
					i |= IS_DESCENDING;
				index_members[index_size++] = i;
				break;
			}
		}
		if (i == cnt) {
			fprintf (stderr, "line %d: unknown field %s in index\n",
				yylineno, member_name);
			cleanup (1);
		}
	} |
	members ',' member
	{
		int	i;
		schema_t *schema = dbm->d_files[file_number].f_schema;
		field_t	*fields = schema->s_fields;
		int	cnt = schema->s_field_count;

		for (i = 1;i < cnt;i++) {
			if (strncmp (member_name, fields[i].f_name,
					MAXFIELDNAME) == 0) {
				if (member_order == L_DESCENDING)
					i |= IS_DESCENDING;
				index_members[index_size++] = i;
				break;
			}
		}
		if (i == cnt) {
			fprintf (stderr, "line %d: unknown field %s in index\n",
				yylineno, member_name);
			cleanup (1);
		}
	}
	;

type:	L_CHAR { field_type = T_CHAR; }
		L_LENGTH L_NUMBER { char_length = atoi (yytext); } |
	L_SHORT { field_type = T_SHORT; } |
	L_INT { field_type = T_INT; } | 
	L_LONG { field_type = T_LONG; } |
	L_DATE { field_type = T_DATE; } |
	L_SERIAL { field_type = T_SERIAL; } |
	L_FLOAT { field_type = T_FLOAT; } |
	L_DOUBLE { field_type = T_DOUBLE; } |
	L_MONEY { field_type = T_MONEY; } |
	L_INDEX
	{
		field_type = T_INDEX;
		index_unique = 0;
	} members |
	L_INDEX error {
		fprintf (stderr, "line %d: expected member list\n",
			yylineno);
		cleanup (1);
	} |
	L_INDEX L_UNIQUE
	{
		field_type = T_INDEX;
		index_unique = 1;
	} members |
	L_INDEX L_UNIQUE error {
		fprintf (stderr, "line %d: expected member list\n",
			yylineno);
		cleanup (1);
	} |
	error {
		fprintf (stderr, "line %d: expected field type, got `%s'\n",
			yylineno, yytext);
		cleanup (1);
	}
	;

field:	L_FIELD L_NAME
	{
		strncpy (field_name, last_name, MAXFIELDNAME);
		field_name[MAXFIELDNAME] = '\0';
	} L_TYPE type {
		field_t	new_field;
		sort_key_t new_key;

		strncpy (new_field.f_name, field_name, MAXFIELDNAME);
		new_field.f_type = field_type;

		if (field_type != T_INDEX) {
			if (field_type == T_CHAR)
				new_field.f_size = char_length;

			if (db_create_field (dbm, file_number,
					&new_field) < 0) {
				fprintf (stderr, "line %d: can't create field %s\n",
					yylineno, field_name);
				db_error (field_name);
				cleanup (1);
			}
		} else {
			new_key.k_unique = index_unique;
			new_key.k_field_count = index_size;
			memcpy (new_key.k_fields, index_members,
				sizeof new_key.k_fields);

			if (db_create_index (dbm, file_number,
					&new_field, &new_key) < 0) {
				fprintf (stderr, "line %d: can't create index %s\n",
					yylineno, field_name);
				db_error (field_name);
				cleanup (1);
			}
		}
	}
	;

fields: field |
	fields field
	;

files:	file fields {
		file_name[0] = '\0';
	} |
	files file fields {
		file_name[0] = '\0';
	}
	;

schema:	database files L_END |
	database error {
		fprintf (stderr, "line %d: expected file definition\n",
			yylineno);
		cleanup (1);
	} |
	database files error {
		fprintf (stderr, "line %d: expected END, got `%s'\n",
			yylineno, yytext);
		cleanup (1);
	}
	;
%%
