/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#ifdef STDLIB_H
#include <stdlib.h>
#endif
#ifdef	MALLOC_H
#include <malloc.h>
#endif
#ifdef BSD
#include <strings.h>
#define strchr index
#define strrchr rindex
#else
#include <string.h>
#ifndef	SYSIII
#include <memory.h>
#endif
#endif /* BSD */
#include "is_types.h"
#include "is_proto.h"

#ifndef lint
static	char	sccsid[] = "@(#)is_open.c	1.8	10:32:13	2/26/92";
#endif

extern	schema_t dict_schema;
extern	is_index_t dict_index;
extern	char	*getenv();
#if !defined(MALLOC_H) && !defined(STDLIB_H)
extern	char	*malloc();
extern	char	*realloc();
#endif

extern	int	db_errno;

int	is_blocksize = MAXBLOCKSIZE;

/*
 * isam_file_open - open an existing ISAM file or create a new one
 *
 * isam_file_open() scans the DBPATH list of directories for the named
 * file (unless the name begins with a "/") to locate the requested
 * file.  If the O_CREAT flag is set, isam_file_open() scans the directory
 * list to see if the file already exists (an error condition) before
 * creating a new file by that name.
 */

file_t *
isam_file_open (name, mode)
char	*name;
int	mode;
{
	off_t	block;
	static	file_t	file;
	char	*cp;
	char	*paths;
	char	new_name[BUFSIZ];
	char	key[BUFSIZ];
	char	data[BUFSIZ];
	header_t header;

	/*
	 * Start with a fresh file structure each time
	 */

#ifdef BSD
	bzero (&file, sizeof file);
#else
	memset (&file, 0, sizeof file);
#endif /* BSD */

	if (strlen (name) > (BUFSIZ-5))
		return (db_errno = DB_INVALID), (file_t *) 0;

	if (mode & ~(O_RDONLY|O_WRONLY|O_RDWR|O_CREAT|O_EXCL))
		return (db_errno = DB_INVALID), (file_t *) 0;

	if ((paths = getenv ("DBPATH")) == (char *) 0)
		paths = ".";

	if (name[0] != '/') {

		/*
		 * Open an already existing database file using
		 * DBPATH to locate the data.
		 */

		do {
			if (cp = strchr (paths, ':')) {
				strncpy (new_name, paths, cp - paths);
				new_name[cp - paths] = '\0';
				cp++;
			} else
				strcpy (new_name, paths);

			strcat (new_name, "/");
			strcat (new_name, name);
			strcat (new_name, ".dat");
		} while (access (new_name, 0) && (paths = cp));

		/*
		 * Construct the absolute pathname for the file by
		 * appending the name to the directory in DBPATH
		 * where the data file was found.
		 */

		if (paths) {
			if (mode & O_CREAT)
				return (db_errno = DB_FILE_EXISTS),
					(file_t *) 0;

			if (cp) {
				strncpy (new_name, paths, cp - paths);
				new_name[cp - paths - 1] = '\0';
			} else
				strcpy (new_name, paths);

			strcat (new_name, "/");
			strcat (new_name, name);
			name = new_name;
		} else if ((mode & O_CREAT) == 0)
			return (db_errno = DB_NOT_FOUND), (file_t *) 0;
	}

	/*
	 * The user asked for a new file.  We know that the file does
	 * not exist in the DBPATH (unless it started with a "/", which
	 * we allow), so we are able to create it.
	 */

	if (mode & O_CREAT) {
		if (name[0] != '/') {

			/*
			 * Pick the first directory in the search path
			 * and use that.
			 */

			if ((paths = getenv ("DBPATH")) == (char *) 0)
				paths = ".";

			if (cp = strchr (paths, ':')) {
				strncpy (new_name, paths, cp - paths);
				new_name[cp - paths] = '\0';
			} else
				strcpy (new_name, paths);

			strcat (new_name, "/");
			strcat (new_name, name);
			name = new_name;
		}

		/*
		 * Append the file extensions to each file to create
		 * the final file names
		 */

		sprintf (key, "%s.idx", name);
		sprintf (data, "%s.dat", name);

		/*
		 * Check the files for existence before actually doing
		 * the create.
		 */

		if (access (data, 0) == 0 || access (key, 0) == 0)
			return (db_errno = DB_FILE_EXISTS), (file_t *) 0;

		if ((file.f_data_fd = open (data, mode, 0666)) == -1)
			return (db_errno = DB_EIO), (file_t *) 0;

		if ((file.f_key_fd = open (key, mode, 0666)) == -1) {
			close (file.f_data_fd);
			return (db_errno = DB_EIO), (file_t *) 0;
		}

		/*
		 * Set up the entire file header.  This describes how
		 * the file was created, where the free list is, etc.
		 * It also provides some magic numbers that get checked
		 * at other times.
		 */

		header.h_free = 0;
		header.h_ticket = file.f_ticket = 1;
		header.h_serial = file.f_serial = 0;
		header.h_bsize = file.f_bsize = is_blocksize;
		header.h_magic = file.f_magic = FILEMAGIC;

		if (write (file.f_key_fd, &header, sizeof header) !=
				sizeof header) {
			db_errno = DB_EIO;
fail:
			close (file.f_data_fd);
			close (file.f_key_fd);
			if (mode & O_CREAT) {
				unlink (key);
				unlink (data);
			}
			return (file_t *) 0;
		}

		/*
		 * Allocate blocks for all the other index root nodes.
		 */

#ifdef	NO_EXTEND
		if (isam_extend (&file, (file.f_bsize * (MAXINDEXFIELDS+1))))
			goto fail;
#endif
		if (lseek (file.f_key_fd, (file.f_bsize *
				(MAXINDEXFIELDS+1)) - sizeof block, 0) == -1) {
			db_errno = DB_EIO;
			goto fail;
		}
		if (write (file.f_key_fd, &block, sizeof block) !=
				sizeof block) {
			db_errno = DB_EIO;
			goto fail;
		}

		/*
		 * Now I do something similiar for the data block free
		 * list.  Having a zero pointer will allocate blocks off
		 * the end of the file.  I don't know how big a record
		 * is though, so I just write out a word.
		 */

		block = 0;
		if (write (file.f_data_fd, &block, sizeof block) !=
				sizeof block) {
			db_errno = DB_EIO;
			goto fail;
		}
	} else {

		/*
		 * Figure out the name of the key and data file and open
		 * both of them.
		 */

		sprintf (key, "%s.idx", name);
		sprintf (data, "%s.dat", name);

		if ((file.f_data_fd = open (data, mode)) == -1) {
			db_errno = DB_EIO;
			return (file_t *) 0;
		}
		if ((file.f_key_fd = open (key, mode)) == -1) {
			close (file.f_data_fd);
			db_errno = DB_EIO;
			return (file_t *) 0;
		}

		/*
		 * Fetch the file header.  It has important information
		 * about this file, like, how big the index blocks are
		 * and so on.
		 */

		if (read (file.f_key_fd, &header, sizeof header) !=
				sizeof header) {
			db_errno = DB_EIO;
			goto fail;
		}

		/*
		 * Update the file structure with the information
		 * from the header.
		 */

		if (header.h_bsize)
			file.f_bsize = header.h_bsize;
		else
			file.f_bsize = MAXBLOCKSIZE;

		/*
		 * The magic number says both what kind of file this
		 * is and what the byte sex is.  This is very important
		 * since the system assumes the index file really is
		 * the right file.  Particularly with data dictionaries
		 * since the schema must be correct or else the system
		 * does strange things.
		 */

		if (header.h_magic) {
			if (header.h_magic != FILEMAGIC &&
					header.h_magic != DICTMAGIC) {
				db_errno = DB_INVALID;
				goto fail;
			}
			file.f_magic = header.h_magic;
		}
	}

	/* 
	 * Allocate space for the index blocks that will be read in.
	 * Index blocks are fixed length, so one size fits all.
	 */

	if ((file.f_key_buffer = (char *) malloc (file.f_bsize)) == 0) {
		db_errno = DB_NOMEM;
		goto fail;
	}

	/*
	 * Find out what my name really is, without all the directory
	 * glarp in front.
	 */

	if (cp = strrchr (name, '/'))
		cp++;
	else
		cp = name;

	strncpy (file.f_file_name, cp, MAXFILENAME);

	/* 
	 * Get the current ticket for the file.  This is used for cache
	 * validation.
	 */

	if ((file.f_ticket = isam_get_ticket (&file)) != -1)
		(void) isam_end_ticket (&file);

	return &file;
}
