#!/bin/sh
#
# Test 7
#	Check for dbschema not corrupting the dictionary
#
#	There are two kinds of errors - ones that cause the entire file
#	to be removed, and ones that leave the new file definition in
#	the dictionary up to a certain point.  This is because of how
#	the grammar is written (and it seemed like a good idea ...)
#
#	@(#)test_07.sh	1.2	09:32:50	3/30/92
#
rm -f test[0-9].dat test[0-9].idx test.out
set -e
RC=1

trap 'rm -f test.out test.sch test_06.o test_06 test[0-9].dat test[0-9].idx ; exit $RC' 0 1 2 15

DBPATH=. ; export DBPATH

#
# create a schema containing 1 field
#

cat > test.sch << EOF
database test0

file test1

field value type int

end
EOF

./dbschema test.sch

#
# get the schema for the new database, and make certain it is correct
#

./dbshow test0 > test.out
cmp - test.out << EOF
database test0

file test1

field	value	type	int

end
EOF

#
# add another file, but make the schema have a syntax error
#

cat > test.sch << EOF
database test0

file test2

field value2 type int
this is a syntax error

end
EOF

#
# if the schema compiles, something is wrong with the compiler!
#
echo 'Expect 2 error messages'
if ./dbschema test.sch; then
	exit $RC
fi

#
# make certain the schema is intact and so on.  this syntax error
# DOES create the new file since the error is at the beginning of
# a field definition.  (long story ...)
#
./dbshow test0 > test.out
cmp - test.out << EOF
database test0

file test1

field	value	type	int

file test2

field	value2	type	int

end
EOF

#
# add another file, but make the schema have a different syntax error
#

cat > test.sch << EOF
database test0

file test3

field value3 syntax error

end
EOF

#
# if the schema compiles, something is wrong with the compiler!
#
echo 'Expect 2 error messages'
if ./dbschema test.sch; then
	exit $RC
fi

#
# make certain the schema is intact and so on.  this syntax error
# DOES NOT create the new file since the error is in the middle of
# a field definition. (another long story)
#
./dbshow test0 > test.out
cmp - test.out << EOF
database test0

file test1

field	value	type	int

file test2

field	value2	type	int

end
EOF

#
# everything is OK, exit
#
RC=0
